#    Copyright 2021 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

import argparse
import os
import sys
import tempfile
import time
import yaml

from si_runtest import logger
from retry import retry

from exec_helpers import SSHClient, SSHAuth, Subprocess
from si_runtest import utils

LOG = logger.logger


EXEC_DEFAULTS = {
    'RUN_ON_REMOTE': False,
    'SEED_SSH_PRIV_KEY_FILE': '',
    'SEED_SSH_PRIV_KEY': '',  # SSH private key content from config file (to write into a tmp file)
    'SEED_STANDALONE_EXTERNAL_IP': '',
    'SEED_SSH_LOGIN': '',
    'VERBOSE': True,
    'SEED_UPLOAD_DIRS': '',
    'SEED_DOWNLOAD_DIRS': '',
    'JOB_NAME': '',
    'BUILD_NUMBER': '',
    'USER': 'si-test',
    'USER_TIME': time.strftime("%Y%m%d_%H-%M-%S"),
    'SEED_WORKSPACE_PATH_PREFIX': '/var/lib/si-tests',
    'SEED_VIRTUALENV_NAME': 'si-venv',
    'SETUP_COMMAND': '',
    'TEARDOWN_COMMAND': '',
    'COMMAND_TIMEOUT': 3600,
    'SETUP_COMMAND_TIMEOUT': 1800,
    'TEARDOWN_COMMAND_TIMEOUT': 1800,
}


CONFIG_CREATE_DEFAULTS = {
    'RUN_ON_REMOTE': False,
    'SEED_SSH_PRIV_KEY_FILE': '',
    'SEED_STANDALONE_EXTERNAL_IP': '',
    'SEED_SSH_LOGIN': 'ubuntu',
    'VERBOSE': True,
    'SEED_UPLOAD_DIRS': '.:.',
    'SEED_DOWNLOAD_DIRS': 'artifacts:artifacts',
    'SETUP_COMMAND': ('[ -d ${SEED_VIRTUALENV} ] || ( sudo apt update; '
                      'sudo apt-get install -y python3.8 python3-virtualenv virtualenv;'
                      ' virtualenv -p python3.8 ${SEED_VIRTUALENV}); '
                      '${SEED_VIRTUALENV}/bin/pip install -r si_tests/requirements.txt; '
                      '${SEED_VIRTUALENV}/bin/pip install -r si_tests/kaas-bm-env/requirements.txt'),
    'TEARDOWN_COMMAND': '',
    'SEED_WORKSPACE_PATH_PREFIX': '/var/lib/si-tests',
    'SEED_VIRTUALENV_NAME': 'si-venv',
}


SKIP_ENV_VARS = [
    'VIRTUAL_ENV',
    'LD_LIBRARY_PATH',
    'DISPLAY',
    'SSH_AGENT_PID',
    'SSH_AUTH_SOCK',
    'HOSTNAME',
    'HOME',
    'JENKINS_HOME',
    'HUDSON_HOME',
    'TMPDIR',
    'WORKSPACE',
    'WORKSPACE_TMP',
    'PWD',
    'PATH',
    'PS1',
]


_boolean_states = {'1': True, 'yes': True, 'true': True, 'on': True,
                   '0': False, 'no': False, 'false': False, 'off': False}


env_var_types_cast = {
    str: lambda name: str(name),
    bool: lambda name: _boolean_states.get(name.lower()) if type(name) is str else bool(name),
    int: lambda name: int(name),
}


def get_exec_defaults(extra_defaults, si_config_path=''):
    """Load defaults

    1. Get defaults from si_config_path YAML key 'run_on_remote'
    2. Overwrite defaults with environment variables if set
    """
    if si_config_path:
        with open(si_config_path, 'r') as f:
            si_config = yaml.load(f.read(), Loader=yaml.SafeLoader)
            config_defaults = si_config.get('run_on_remote', {})
    else:
        config_defaults = {}

    for key, value in extra_defaults.items():
        value_type = type(value) if type(value) in env_var_types_cast else str
        env_value = os.environ.get(key) or config_defaults.get(key, value)
        env_value = env_var_types_cast[value_type](env_value)
        config_defaults[key] = env_value

    return config_defaults


@retry(Exception, delay=1, tries=3, jitter=1, logger=LOG)
def remote_seed(opts):
    """Init an SSHClient object"""
    key = utils.load_keyfile(opts.seed_ssh_priv_key_file)
    key = utils.get_rsa_key(key["private"])
    seed_ip = opts.seed_standalone_external_ip
    assert seed_ip, 'SEED_STANDALONE_EXTERNAL_IP is empty!'
    remote = SSHClient(
        host=seed_ip,
        port=22,
        auth=SSHAuth(
            username=opts.seed_ssh_login,
            key=key))
    remote.logger.addHandler(logger.console)
    return remote


def rsync_run(opts, verbose=False, reverse=False,
              from_dir=None, todir=None, from_file=None, tofile=None):
    """
    Run local rsync call
    :param username:
    :param hostname:
    :param sshkeyfile:
    :param from_dir:
    :param todir:
    :param from_file:
    :param tofile:
    :param openstack_client:
    :param reverse: False -> sync from local to host
                    True -> sync from host to local
    :return:
    """
    hostname = opts.seed_standalone_external_ip
    username = opts.seed_ssh_login
    sshkeyfile = opts.seed_ssh_priv_key_file

    # we need to full path to key,in case need to run from pushd
    sshopts = (f'"ssh -i'
               f' {os.path.abspath(sshkeyfile)}'
               ' -o StrictHostKeyChecking=no'
               ' -o ConnectionAttempts=10 '
               ' -o UserKnownHostsFile=/dev/null "')
    # 'si-runtest.log' - reserved name for the log file
    exclude_opts = ("--exclude 'si-runtest.log'")
    if not (todir or tofile):
        raise Exception("Target not set")
    r = Subprocess()
    r.logger.addHandler(logger.console)
    if from_file:
        _to = tofile or todir
        if reverse:
            LOG.info(f"Downloading '{hostname}:{from_file}' to '{_to}'")
            cmd = f'rsync -e {sshopts} {exclude_opts} -azP {username}@{hostname}:{from_file} {_to}'
            r.check_call(cmd, raise_on_err=True, verbose=verbose, timeout=3600)
        else:
            cmd = f'rsync -e {sshopts} {exclude_opts} -azP {from_file} {username}@{hostname}:{_to}'
            LOG.info(f"Uploading '{from_file}' to '{hostname}:{_to}'")
            r.check_call(cmd, raise_on_err=True, verbose=verbose, timeout=3600)
    elif from_dir:
        if reverse:
            from_dir = from_dir + '/'
            cmd = f'rsync -e {sshopts} {exclude_opts} -azP {username}@{hostname}:{from_dir} .'
            with utils.pushd(todir):
                LOG.info(f"Downloading '{hostname}:{from_dir}' dir to '{todir}'")
                r.check_call(cmd, raise_on_err=True, verbose=verbose, timeout=3600)
        else:
            cmd = f'rsync -e {sshopts} {exclude_opts} -azP . {username}@{hostname}:{todir}'
            with utils.pushd(from_dir):
                LOG.info(f"Uploading '{from_dir}' dir to '{hostname}:{todir}'")
                r.check_call(cmd, raise_on_err=True, verbose=verbose, timeout=3600)
    else:
        raise Exception('Wrong invocation')


def create_workspace(opts):
    """Create a workspace directory with access rights for the current user"""
    remote = remote_seed(opts)
    path = opts.seed_workspace_path_prefix
    name = opts.seed_workspace_name
    LOG.info(f"Creating workspace at {path}/{name}")
    cmd = f"sudo mkdir -p {path}/{name} && sudo chown -R $(id -u):$(id -g) {path}"
    remote.check_call(command=cmd, verbose=False, timeout=60)


def dirs(paths):
    """Generate pairs of source:destination pairs from the paths string

    :param paths: semicolon-separated pairs of source:destination paths
                  separated with a colon. For example:
                  '.:.;~/.cache/pip:~/.cache/pip'
    """
    pairs = paths.split(';')
    for pair in pairs:
        if ':' in pair:
            yield pair.split(":")
        else:
            LOG.warning(f"Wrong path format, no colon found: '{pair}'")


def upload_dirs(opts):
    """Upload directories specified in opts.seed_upload_dirs to the seed node"""
    path = opts.seed_workspace_path_prefix
    name = opts.seed_workspace_name
    for source, destination in dirs(opts.seed_upload_dirs):
        if not destination.startswith("~") and not destination.startswith("/"):
            destination = f"{path}/{name}/{destination}"
        rsync_run(opts, reverse=False, from_dir=source, todir=destination)


def download_dirs(opts):
    """Download directories specified in opts.seed_download_dirs from the seed node"""
    path = opts.seed_workspace_path_prefix
    name = opts.seed_workspace_name
    for source, destination in dirs(opts.seed_download_dirs):
        if not source.startswith("~") and not source.startswith("/"):
            source = f"{path}/{name}/{source}"
        rsync_run(opts, reverse=True, from_dir=source, todir=destination)


def get_env_vars(opts):
    """Make a list of environment variables to expose into remote node

    - Skip the environment variables that match the already existing
      variables on the remote.
    - Skip the environment variables that match the SKIP_ENV_VARS list.
    - Add extra environment variables WORKSPACE and SEED_VIRTUALENV
    """
    cwd = os.getcwd()
    remote = remote_seed(opts)
    cmd = 'env'
    ret = remote.check_call(command=cmd, verbose=False, timeout=60)
    remote_env_vars = [line.split("=")[0] for line in ret.stdout_str.splitlines()]

    local_env_vars = {key: val
                      for key, val in os.environ.items()
                      if key not in remote_env_vars and key not in SKIP_ENV_VARS}
    local_env_vars['WORKSPACE'] = f"{opts.seed_workspace_path_prefix}/{opts.seed_workspace_name}"
    local_env_vars['SEED_VIRTUALENV'] = f"{opts.seed_workspace_path_prefix}/{opts.seed_virtualenv_name}"
    LOG.debug(yaml.dump(local_env_vars))

    cwd = os.getcwd()
    for key, value in local_env_vars.items():
        if value.startswith(cwd):
            LOG.warning(f"=== Environment variable {key} provides a local path"
                        f" that cannot be used on the remote node: {value}")
    return local_env_vars


def create_remote_source_file(opts, env_vars, workspace, seed_virtualenv):
    """Create a source file on the remote node inside workspace directory

    In additional, try to activate the python virtualenv, if it was created
    in the custom --setup-command
    """
    remote = remote_seed(opts)
    env_vars_str = '\n'.join([f"export {key}='{val}'" for key, val in env_vars.items()])
    source_file = f'{workspace}/.env-si-test'
    cmd = (f"cat << 'EOF' > {source_file}\n"
           f"{env_vars_str}\n"
           f"[ -f {seed_virtualenv}/bin/activate ] && . {seed_virtualenv}/bin/activate ||"
           f" echo WARNING: python virtualenv not found at {seed_virtualenv}\n"
           f"EOF\n")
    remote.check_call(command=cmd, verbose=False, timeout=60)
    return source_file


def runcmd_remote_quiet(cmd, opts, timeout):
    """Execute the command with disabled verbose

    Use to prepare the environment and cleanup steps"""
    remote = remote_seed(opts)
    if cmd:
        LOG.info(f"Quiet run the command: '{cmd}'")
        remote.check_call(command=cmd, verbose=False, timeout=timeout)


def runcmd_remote(opts):
    """Executing the command remotelly

    1. Create an unique workspace on remove
    2. Rsync the current directory to the remote workspace (path '.' must be in SEED_UPLOAD_DIRS)
    3. Rsync the pip cache to the remote userdir cache (path to cache must be in SEED_UPLOAD_DIRS)
    4. Create/update the remote virtualenv (commands must be in SETUP_COMMAND)
    5. Install requirements into the remote virtualenv (commands must be in SETUP_COMMAND)
    6. Activate the remote virtualenv and run the main command
    7. Execute cleanup steps (commands must be in TEARDOWN_COMMAND)
    8. Rsync the 'artifacts' directory from the remote workspace (paths must be in SEED_DOWNLOAD_DIRS)
    """
    workspace = f"{opts.seed_workspace_path_prefix}/{opts.seed_workspace_name}"
    seed_virtualenv = f"{opts.seed_workspace_path_prefix}/{opts.seed_virtualenv_name}"

    create_workspace(opts)
    upload_dirs(opts)

    env_vars = get_env_vars(opts)
    source_file = create_remote_source_file(opts, env_vars, workspace, seed_virtualenv)

    setup_command = f"cd {workspace}; . {source_file}; {opts.setup_command}"
    teardown_command = f"cd {workspace}; . {source_file}; {opts.teardown_command}"
    main_command = f"cd {workspace}; . {source_file}; {opts.command}"

    runcmd_remote_quiet(setup_command, opts, opts.setup_command_timeout)
    try:
        verbose = opts.verbose
        remote = remote_seed(opts)
        ret = remote.execute(command=main_command, verbose=verbose, timeout=opts.command_timeout)
    finally:
        time.sleep(5)  # To print out the logs from the remote stream
        download_dirs(opts)
        runcmd_remote_quiet(teardown_command, opts, opts.teardown_command_timeout)

    return ret


def runcmd_local(opts):
    """Executing the command locally"""

    cmd = opts.command
    verbose = opts.verbose

    subprocess = Subprocess()
    subprocess.logger.addHandler(logger.console)
    ret = subprocess.execute(command=cmd, verbose=verbose, timeout=opts.command_timeout)
    return ret


def load_params(args, defaults=None):
    """
    Parse CLI arguments and environment variables

    Returns: ArgumentParser instance
    """
    defaults = defaults or {}
    if defaults.get('JOB_NAME') and defaults.get('BUILD_NUMBER'):
        default_workspace_path = f"workspace_{defaults.get('JOB_NAME')}_{defaults.get('BUILD_NUMBER')}"
    else:
        default_workspace_path = f"workspace_{defaults.get('USER')}_{defaults.get('USER_TIME')}"

    common_parser = argparse.ArgumentParser(add_help=False)
    common_parser.add_argument('--run-on-remote', dest='run_on_remote',
                               action='store_const', const=True,
                               help='Run the command on a remote node instead of the current local directory',
                               default=defaults.get('RUN_ON_REMOTE', False))
    common_parser.add_argument('--seed-ssh-priv-key-file',
                               help='Path to the SSH private key file to access the seed node',
                               default=defaults.get('SEED_SSH_PRIV_KEY_FILE', ''),
                               type=str)
    common_parser.add_argument('--seed-standalone-external-ip',
                               help='Seed node IP address',
                               default=defaults.get('SEED_STANDALONE_EXTERNAL_IP', ''),
                               type=str)
    common_parser.add_argument('--seed-ssh-login',
                               help='Seed node SSH username',
                               default=defaults.get('SEED_SSH_LOGIN', ''),
                               type=str)
    common_parser.add_argument('--seed-upload-dirs',
                               help='Upload directories to seed node before run tests',
                               default=defaults.get('SEED_UPLOAD_DIRS', ''),
                               type=str)
    common_parser.add_argument('--seed-download-dirs',
                               help='Download directories from seed node after run tests',
                               default=defaults.get('SEED_DOWNLOAD_DIRS', ''),
                               type=str)
    common_parser.add_argument('--seed-workspace-path-prefix',
                               help='Path to the directory for workspaces from different tests',
                               default=defaults.get('SEED_WORKSPACE_PATH_PREFIX'),
                               type=str)
    common_parser.add_argument('--seed-virtualenv-name',
                               help=('Name of the python virtualenv on the seed node that will be activated '
                                     'to run the command, if exists.\n'
                                     'If required, please init the virtualenv using --setup-command with '
                                     '${SEED_VIRTUALENV} path'),
                               default=defaults.get('SEED_VIRTUALENV_NAME'),
                               type=str)
    common_parser.add_argument('--setup-command',
                               help=('Command to execute inside the created workspace before run the main command,'
                                     ' with suppressed verbose'),
                               default=defaults.get('SETUP_COMMAND', ''),
                               type=str)
    common_parser.add_argument('--teardown-command',
                               help=('Command to execute inside the created workspace after run the main command,'
                                     ' with suppressed verbose'),
                               default=defaults.get('TEARDOWN_COMMAND', ''),
                               type=str)
    common_parser.add_argument('--command-timeout',
                               help=('Main command timeout'),
                               default=defaults.get('COMMAND_TIMEOUT', 3600),
                               type=int)
    common_parser.add_argument('--setup-command-timeout',
                               help=('Setup command timeout'),
                               default=defaults.get('SETUP_COMMAND_TIMEOUT', 1800),
                               type=int)
    common_parser.add_argument('--teardown-command-timeout',
                               help=('Teardown command timeout'),
                               default=defaults.get('TEARDOWN_COMMAND_TIMEOUT', 1800),
                               type=int)

    exec_parser = argparse.ArgumentParser(add_help=False)
    exec_parser.add_argument('--command',
                             help='Main command to execute inside the created workspace,'
                                  ' usually run pytest with a test',
                             default='',
                             type=str)
    exec_parser.add_argument('--si-config-path',
                             help='(Optional) Config with extra options for test runner',
                             default=defaults.get('SI_CONFIG_PATH', ''),
                             type=str)
    exec_parser.add_argument('--seed-workspace-name',
                             help=('Name of the current workspace on the seed node\n'
                                   'Default: "workspace_${JOB_NAME}_${BUILD_NUMBER}" if set,'
                                   ' else "workspace_${USER}_${USER_TIME}" , where USER_TIME is YYYYMMDD_HH-MM-SS"\n'
                                   'The environment variable ${WORKSPACE} with the full path to the workspace'
                                   'will be available for --setup-command and --teardown-command'),
                             default=default_workspace_path,
                             type=str)

    verbose_parser = argparse.ArgumentParser(add_help=False)
    verbose_parser.add_argument('--verbose', dest='verbose',
                                action='store_const', const=True,
                                help='Show verbosed output',
                                default=defaults.get('VERBOSE', False))

    parser = argparse.ArgumentParser(
        description="Execute a shell command locally or on a remote host.\n"
                    "For additional help, use with -h/--help option")
    subparsers = parser.add_subparsers(title="Options",
                                       help='Available options',
                                       dest='option')

    subparsers.add_parser('exec',
                          parents=[exec_parser, common_parser, verbose_parser],
                          help="Execute command")

    subparsers.add_parser('config-create',
                          parents=[common_parser, verbose_parser],
                          help="Render a si-config YAML to stdout from defaults and provided arguments")

    if len(args) == 0:
        args = ['-h']
    return parser.parse_args(args)


def option_exec(opts, args):

    # Get the --si-config argument and load default values from the config
    # Priorities for the arguments, from high to low:
    # - arguments from CLI
    # - arguments from environment variables
    # - arguments from --si-config YAML file
    # - defaults defined in DEFAULTS variable
    defaults = get_exec_defaults(EXEC_DEFAULTS, opts.si_config_path)
    # Get arguments from CLI, using extra defaults
    opts = load_params(args, defaults)

    if not opts.command:
        LOG.info("--command argument not set")
        return 11

    LOG.info(f"=== RUN_ON_REMOTE: {opts.run_on_remote} ===")
    if opts.run_on_remote:
        if opts.seed_ssh_priv_key_file:
            ret = runcmd_remote(opts)
        elif defaults['SEED_SSH_PRIV_KEY']:
            with tempfile.NamedTemporaryFile(mode="w") as tmp:
                tmp.write(defaults['SEED_SSH_PRIV_KEY'])
                tmp.flush()
                opts.seed_ssh_priv_key_file = tmp.name
                ret = runcmd_remote(opts)
        else:
            raise Exception("SEED_SSH_PRIV_KEY_FILE not set")
    else:
        ret = runcmd_local(opts)

    return ret.exit_code


def option_config_create(opts, args):
    defaults = get_exec_defaults(CONFIG_CREATE_DEFAULTS)
    opts = load_params(args, defaults)
    # Log informational messages to stderr, to not break the YAML structure in stdout
    LOG.error("=== Render si-config YAML to stdout")

    seed_ssh_priv_key = ''
    if opts.seed_ssh_priv_key_file and os.path.isfile(opts.seed_ssh_priv_key_file):
        with open(opts.seed_ssh_priv_key_file, 'r') as f:
            seed_ssh_priv_key = f.read()
    elif opts.run_on_remote:
        raise Exception("SEED_SSH_PRIV_KEY_FILE not set or file doesn't exist")

    config = {
        'run_on_remote': {
            'RUN_ON_REMOTE': opts.run_on_remote,
            'SEED_STANDALONE_EXTERNAL_IP': opts.seed_standalone_external_ip,
            'SEED_SSH_LOGIN': opts.seed_ssh_login,
            'SEED_UPLOAD_DIRS': opts.seed_upload_dirs,
            'SEED_DOWNLOAD_DIRS': opts.seed_download_dirs,
            'SETUP_COMMAND': opts.setup_command,
            'TEARDOWN_COMMAND': opts.teardown_command,
            'SEED_WORKSPACE_PATH_PREFIX': opts.seed_workspace_path_prefix,
            'SEED_VIRTUALENV_NAME': opts.seed_virtualenv_name,
            'SEED_SSH_PRIV_KEY': seed_ssh_priv_key,
            'VERBOSE': opts.verbose,
        }
    }

    def str_presenter(dumper, data):
        if len(data.splitlines()) > 1:
            return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
        return dumper.represent_scalar('tag:yaml.org,2002:str', data)

    yaml.add_representer(str, str_presenter)

    print(yaml.dump(config))


OPTIONS = {
    'exec': option_exec,
    'config-create': option_config_create,
}


def main():
    args = sys.argv[1:]

    opts = load_params(args)
    option = opts.option

    if option not in OPTIONS.keys():
        print(f"Unsupported option {option}, please use one of: "
              f"{OPTIONS.keys()}")
        return 12

    method = OPTIONS[option]
    exit_code = method(opts, args)

    return exit_code


if __name__ == '__main__':
    sys.exit(main())
